/*
 * @(#)PrologInterface.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;
	using System.Reflection;

	public sealed class PrologInterface {
		private bool result;
		private bool stopping = false;
		private Predicate code = null;
		public Prolog engine;

		private bool daemon = false;
		private bool signalExit = false;

		public PrologInterface( ) {
			engine = new Prolog( );
			daemon = true;
			signalExit = true;
		}

		public PrologInterface( PushbackReader new_in, 
			System.IO.TextWriter new_out,
			System.IO.TextWriter new_err ) {
			engine = new Prolog( new_in, new_out, new_err );
			daemon = true;
			signalExit = true;
		}

		public PrologInterface( bool daemon ) {
			engine = new Prolog( );
			this.daemon = daemon;
			this.signalExit = true;
		}

		public PrologInterface( PushbackReader new_in, 
			System.IO.TextWriter new_out,
			System.IO.TextWriter new_err, 
			bool daemon ) {
			engine = new Prolog( new_in, new_out, new_err );
			this.daemon = daemon;
			this.signalExit = true;
		}

		internal PrologInterface( bool daemon, bool signalExit ) {
			engine = new Prolog( );
			this.daemon = daemon;
			this.signalExit = signalExit;
		}

		internal PrologInterface( PushbackReader new_in, 
			System.IO.TextWriter new_out,
			System.IO.TextWriter new_err, bool deamon, bool signalExit ) {
			engine = new Prolog( new_in, new_out, new_err );
			this.daemon = daemon;
			this.signalExit = signalExit;
		}

		public void SetPredicate(Predicate p) {
			code = p;
		}

		[Obsolete( "use SetPredicate instead" )] 
		public void setPredicate( Predicate p ) { SetPredicate( p ); }
		[Obsolete( "use Call instead" ) ] 
		public bool call() { return Call(); }
		[Obsolete( "use Redo instead" ) ] 
		public bool redo() { return Redo(); }
		[Obsolete( "use Success instead" ) ] 
		public void success() { Success(); }
		[Obsolete( "use Fail instead" ) ] 
		public void fail() { Fail(); }
		[Obsolete( "use Stop instead" ) ] 
		public void stop() { Stop(); }
		[Obsolete( "use AddCallingAssembly instead" ) ] 
		public void addCallingAssembly() { AddCallingAssembly(); }
		[Obsolete( "use AddAssembly instead" ) ] 
		public void addAssembly( Assembly a ) { AddAssembly( a ); }
		
		public bool Call() {
			lock( this ) {
				this.Start();	 /* execute Run() in new thread */
				try {
					Monitor.Wait(this);	     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}

		public bool Redo() {
			lock( this ) {
				Monitor.Pulse(this);	    /* resume child thread */
				try {
					Monitor.Wait(this);     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}


		public void Success() {
			lock( this ) {					   
				result = true;
				Monitor.Pulse(this);   	/* resume caller's thread */
				try {
					Monitor.Wait(this); 	/* wait child thread */
				} catch (ThreadInterruptedException) {}
			}
		}

		public void Fail() {
			lock( this ) {
				result = false;
				Monitor.Pulse(this);	    /* resume caller's thread */
			}
		}

		private Thread t = null;

		internal void Start() {
			t = new Thread( new ThreadStart( Run ) );
			if( daemon )
				t.IsBackground = true;
			t.Start( );
		}

		public void Stop() {
			stopping = true;
		}

		public void AddAssembly( Assembly ass ) {
			AssemblyList.Add( ass, null );
		}

		public void AddCallingAssembly( ) {
			if( !Assembly.GetCallingAssembly( ).Equals( Assembly.GetExecutingAssembly( ) ) )
				AssemblyList.Add( Assembly.GetCallingAssembly( ), null );
		}

		internal void Run() {
			try {
				try {
					while (true) {
						if (code == null || stopping) 
							break;
						code = code.exec( engine );
						code = engine.ServiceQueue( code );
					}
				}
				catch (AbortException e){
					System.Console.WriteLine(e.Message);
				}
				Fail();
				/* START JJC */
				if( signalExit )
					foreach( ConcurrentVariable cv in engine.concurrentVars )
						cv.ReportThreadDeath( engine );
				/* END JJC */
			} finally {
				engine.ReleaseAllLocks( );
			}
		}
	}
}