/*
 * @(#)VariableTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;

	[System.Serializable]
	public sealed class VariableTerm : Term {

		private static long ctr = Int64.MinValue;
				
		private static long IncCounter( ) {
			lock( typeof( VariableTerm ) ) {
				return ctr++;
			}
		}

		/* START JJC - Added for fork/wait support */

		internal ConcurrentVariable concurrentRoot = null;
		Prolog engine = null;

		// added to support lock/unlock.

		internal object locked_object = null;
		internal object unlocked_object = null;

		/* END JJC */

		Term val;

		private long itimeStamp;
		public long timeStamp {
			get {
				return itimeStamp;
			}
			set {
				itimeStamp = value;
			}
		}
		private long ID;

		public VariableTerm(long t, Prolog engine) {
			this.engine = engine;
			val = this;
			timeStamp = t;
			ID = IncCounter( );
		}

		public VariableTerm() {
			val = this;
			timeStamp = Int64.MinValue;
			ID = IncCounter( ); 
		}

		public override Term Dereference() {
			if (val == this)
				return this;
			return val.Dereference();
		}

		public void Undo() { 
			val = this;
			if( locked_object != null ) {
				System.Threading.Monitor.Exit( locked_object );
				engine.ReportUnlock( locked_object );
				locked_object = null;
			}
			if( unlocked_object != null ) {
				System.Threading.Monitor.Enter( unlocked_object );
				engine.ReportLock( unlocked_object );
				unlocked_object = null;
			}
		}

		public override Term CopyTerm(Prolog engine) {
			System.Collections.Hashtable table = engine.varTable;
			VariableTerm copy;
			if (val != this)
				return val.CopyTerm(engine);
			copy = (VariableTerm)table[this];
			if ( copy == null ){
				copy = engine.makeVariable();
				table[ this ] = copy;
			}
			return copy;
		}

		/* START JJC */

		internal override Term CloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			System.Collections.Hashtable cloneTable = oldEngine.cloneVarTable;
			if (val != this)
				return val.CloneTerm( oldEngine, newEngine, makeConc );

			VariableTerm clone = (VariableTerm)cloneTable[ this ]; // newEngine.makeVariable( );

			if ( clone == null ){ 
				clone = newEngine.makeVariable();
				cloneTable[ this ] = clone;
			}

			// copy the root to the new clone.
			if( makeConc ) {
				if( concurrentRoot == null ) {
					concurrentRoot = new ConcurrentVariable( );
					concurrentRoot.ReportFork( oldEngine );
					oldEngine.concurrentVars.Add( concurrentRoot );
				}
				clone.concurrentRoot = this.concurrentRoot;
				concurrentRoot.ReportFork( newEngine );
				newEngine.concurrentVars.Add( concurrentRoot );
			}
			
			return clone;
		}
		/* END JJC */

		public bool OldUnify( Term t, Trail trail ) {
			VariableTerm v1;
			if (val != this)
				return val.Unify(t, trail);
			
			t = t.Dereference();
			if ( t.IsVariable() ){
				VariableTerm v2 = (VariableTerm)t;
				if (this.timeStamp < v2.timeStamp ){
					v2.val = this;
					v1 = v2;
				} else {
					this.val = t;
					v1 = this;
				}
			} else {
				this.val = t; 
				v1 = this;
			}
			if ( v1.timeStamp < trail.engine.stack.GetTimeStamp() ){
				trail.Push(v1);
			}
			return true;
		}

		public override bool Unify(Term t, Trail trail) {
			VariableTerm v1;
			if (val != this)
				return val.Unify(t, trail);
			
			t = t.Dereference();
			if ( t.IsVariable() ){
				VariableTerm v2 = (VariableTerm)t;
				if (this.timeStamp < v2.timeStamp ){
					v2.val = this;

					/* START JJC */
					if( this.concurrentRoot != null )
						v2.concurrentRoot = this.concurrentRoot;
					/* END JJC */

					v1 = v2;
				} else {
					this.val = t; 
					
					/* START JJC */					
					if( v2.concurrentRoot != null )
						this.concurrentRoot = v2.concurrentRoot;
					/* END JJC */

					v1 = this;
				}
			} else {
				this.val = t;
				if( this.concurrentRoot != null ) {
					if( trail.engine.concInstQueue == null )
						trail.engine.concInstQueue = new System.Collections.Queue( );
					trail.engine.concInstQueue.Enqueue( 
						new NewBinding( this.concurrentRoot.csObject, t ) );
				}
				v1 = this;
			}
			if ( v1.timeStamp < trail.engine.stack.GetTimeStamp() ){
				trail.Push(v1);
			}
			return true;
		}

		public override bool Equals(Term t) {
			if( val != this )
				return val.Equals(t);
			if ( !t.IsVariable() )
				return false;
			return this == t;
		}

		public override object ToCsObject() {
			if (val != this)
				return val.ToCsObject();
			return this;
		}

		public String Name() {
			return "_" + this.GetHashCode( ).ToString( );
		}

		public long GetID() {
			return ID;
		}

		public override String ToString() {
			if (val != this)
				return val.ToString();
			return Name();
		}

		/* START JJC (for interop from C# side) */

		public void Send( Term inst ) {
			PrologInterface pi = new PrologInterface( true, false );
			this.Undo( );
			pi.SetPredicate( new Predicates.dollar_unify_2( this, inst, null ) );
			pi.Run( );
		}

		public Term Receive( ) {
			PrologInterface pi = new PrologInterface( true, false );
			this.Undo( );
			pi.SetPredicate( new Predicates.WaitForOne_1( this, null ) );
			pi.Run( );
			return this;
		}

		/* END JJC */

	}
}


